import nani


class Bool(nani.Bool):
    __slots__ = ()


class Object(nani.Object):
    __slots__ = ()


class Number(nani.Number):
    __slots__ = ()


class Array(nani.Array):
    __slots__ = ()


class Structure(nani.Structure):
    __slots__ = ()
