import numpy


class Bool(int):
    pass


class Int(int):
    pass


class Float(float):
    pass


class Complex(complex):
    pass


class NumpyInt(numpy.int32):
    pass


class NumpyFloat(numpy.float32):
    pass


class NumpyComplex(numpy.complex64):
    pass
